/*
 * Copyright (c) 2008, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
 * may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <stdint.h>
#ifndef WIN32
#include <libgen.h>
#endif
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <cctype>
#include <cstdio>
#include <iostream>
#include <cstring>
#include <string>
#include <cstdlib>

#include "../logic/defines.h"
#include "../logic/util.h"
#include "../logic/dict.h"
#include "../logic/solver.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/score.h"

#ifdef WIN32
# define random rand
# define srandom srand
#endif

using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::string;
using namespace Spellathon;

static void	solveit(const char *letters, const char *dpath);
static void	playit(const char *dpath, int min, int max);
static void	playit_precomp(const char *dpath, int minsol, int maxsol);

static void	print_game(const char *word);
static void	play_loop(const char *w, Solver *s);
static int	sidxcmp(const void *a, const void *b);

static void	help(void);
static void	instr(void);

bool	keepscore = true;

int
main(int argc, char *argv[])
{
	char	letters[MAX_LETRS+1];
	char	dictpath[PATH_MAX];
	char	*tok = 0;
	int	i, ch, dflag=0, sflag=0, pflag=0, cflag=0;
	int 	minsol = MIN_SOL, maxsol = MAX_SOL;

#ifndef WIN32
	chdir(dirname(argv[0]));
#endif

	if (argc < 2)
		help();
	while ((ch = getopt(argc, argv, "nhcid:pr:s:")) != -1) {
		switch (ch) {
		case 'h':
			help();
			break;

		case 'i':
			instr();
			break;

		case 'r':
			minsol = maxsol = 0;
			minsol = strtol(optarg, 0, 10);
			if (minsol > 0) tok = strchr(optarg, ',');
			if (tok)	maxsol = strtol(tok+1, 0, 10);
			if (!maxsol || maxsol - minsol < MGAP_SOL) {
				cerr << "Invalid solution range."
					" Compiled limit: " 
				     << MMIN_SOL << "," << MMAX_SOL << endl;
				exit(1);
			}
			break;

		case 's':
			sflag++;
			strncpy(letters, optarg, MAX_LETRS);
			letters[MAX_LETRS] = 0;
			for (i = 0; i < MAX_LETRS; i++) {
				if (!isalpha(letters[i])) {
					cerr << "Please enter seven letters.\n";
					exit(1);
				}
				letters[i] = tolower(letters[i]);
			}
			swapch(&letters[0], &letters[MAX_LETRS-1]);
			break;

		case 'p':
			pflag++;
			break;

		case 'c':
			cflag++;
			break;

		case 'd':
			dflag++;
			strncpy(dictpath, optarg, sizeof(dictpath)-1);
			dictpath[sizeof(dictpath)-1] = 0;
			break;

		case 'n':
			keepscore = false;
			break;

		case '?':
		default:
			help();
			break;
		}
	} 
	argc -= optind;
	argv += optind;

	if (!dflag) 
		strcpy(dictpath, DICTPATH);

	if (sflag && pflag) 
		help();

	if (sflag) 
		solveit(letters, dictpath);

	if (pflag) {
		srandom(time(0));

		if (cflag)
			playit_precomp(dictpath, minsol, maxsol);
		else
			playit(dictpath, minsol, maxsol);
	}

	return 0;
}

static int 
sidxcmp(const void *a, const void *b) {
        unsigned int (*c)[2] = (unsigned int (*)[2])a;
        unsigned int (*d)[2] = (unsigned int (*)[2])b;

        if ((*c)[1] == (*d)[1])
                return 0;
        if ((*c)[1] < (*d)[1])
                return -1;
        else
                return 1;
}

static void
playit_precomp(const char *dpath, int minsol, int maxsol)
{
	Dict *d;
	Solver *sol;
	try {
		d = new Dict(dpath, true);
		sol = new Solver(d);
	}
	catch (string s)
	{
		cerr << s << endl;
		exit(1);
	}

	int i = 0, s = -1, e = -1;

	while (i < d->pc_nent()) {
		if (d->precomp_p[i].nsol >= minsol) {
			s = i;
			break;
		}
		i++;
	}

	while (i < d->pc_nent()) {
		if (d->precomp_p[i].nsol > maxsol) {
			break;
		}
		i++;
	}

	e = i;

	if (s < 0) {
		cout << "\rNo suitable game found.\n"
			"Choose a different Solution range"
			" or use a different dictionary.\n";
	} else {
		int r = randnum(s, e);
		sol->solve(d->precomp_p[r].sign);
		play_loop(d->precomp_p[r].sign, sol);
	}

	delete d;
	delete sol;
}
	
static void
playit(const char *dpath, int minsol, int maxsol)
{
	char		sig[MAX_LETRS+1];
	unsigned int 	(*sidx)[2];
	enum	{IDX, RAND};

	Dict *d;
	Solver *sol;
	try {
		d = new Dict(dpath, false);
		sol = new Solver(d);
	}
	catch (string s)
	{
		cerr << s << endl;
		exit(1);
	}

        sidx = (unsigned int (*)[2])memalloc(2 * d->nent7() * sizeof(int));

	/* shuffle the indexes. Knuth shuffle? */
	int j = d->off7();
        for (int i = 0; i < d->nent7(); i++) {
                sidx[i][IDX] = j++;
                sidx[i][RAND] = random();
        }

        qsort((void*)sidx, d->nent7(), 2 * sizeof(int), sidxcmp);

	unsigned int cidx[MAX_LETRS][2];
        for (int i = 0; i < MAX_LETRS; i++) {
                cidx[i][IDX] = i;
                cidx[i][RAND] = random();
        }

        qsort((void*)cidx, MAX_LETRS, 2 * sizeof(int), sidxcmp);

	/*
	 * find a random seven letter signature from the dict and solve
	 * it and any of the six different central letter combinations.
	 * consider the word playable if MIN_SOL <= solutions <= MAX_SOL
	 *
	 * this technique works well. it's fast, sufficiently
	 * random and covers all the signatures.
	 */
	for (int i = 0; i < d->nent7(); i++) {

		const char *dsig = d->dp[sidx[i][IDX]].sign;

		int j;
		for (j = 0 ; j < MAX_LETRS; j++)
			sig[j] = dsig[cidx[j][IDX]]; 

		sig[j] = 0;

		int k = MAX_LETRS - 2;
		do {
			if (sol->solve(sig, minsol, maxsol) != -1) {
				/* found suitable word */
				free(sidx);
				play_loop(sig, sol);
			}
			if (k < 0)
				break;

			/* try a different central letter */	
			swapch(&sig[MAX_LETRS-1], &sig[k--]);

		} while (1);

		/* 
		 * this can take some time in pathological cases 
		 * so we need to keep the gui responsive 
		 */	
		if (!(i & ((1<<8) - 1))) 
			spinner();
	}

	free(sidx);
	cout << "\rNo suitable game found.\nChoose a different Solution range"
			      " or use a different dictionary.\n";

	delete d;
	delete sol;
}

void
play_loop(const char *w, Solver *s)
{
	DispWordsBuf dsp;
	dsp.init(s->words());
	time_t t = time(0);
	string op = "";

	while (1) {
		print_game(w);
		cout << dsp.hintBuf() << endl;
		cout << "words: " << dsp.nSetWords() << "/" << dsp.nWords() 
		     << "        ";
		cout << op << "\n\n";
		cout << "Enter a word ('f' to finish, 'h' for hint) : ";
		string word;
		cin >> word;

		if (word == "h") {
			if (!dsp.setHint()) 
				op = "< No more hints >";

			continue;
		}

		if (word == "f") {
fin:
			time_t t2 = time(0);

			cout << endl << dsp.solBuf() << endl;
			if (keepscore) {
				cout << "< Words: " << dsp.nSetWords() 
				     << " out of " << dsp.nWords() <<
				     ".    Time: " << t2-t << " secs. >\n";

				Score sc(dsp.nSetWords(), dsp.nWords(), 
					 dsp.hintLevel(), t2-t);

				cout << "< Score: " << (int)sc.score() <<
				     "     Rating: " << sc.rating() << " >\n\n"; 
			} else {
				cout << "< Finished in " << t2-t 
				     << " secs. >\n\n";
			}

			exit(0);
		}

		if (word.size() > 7 || word.size() < 4) {
				op = "< No match >";
		}

		switch (dsp.setWord(word)) {
		case 1:
			op = "< Correct >";
			if (dsp.nWords() == dsp.nSetWords()) 
				goto fin;			
			break;
		case -1:
			op = "< No match >";
			break;
		case 0:
			op = "< Already entered >";
			break;
		}
	}
}

void
print_game(const char *word)
{
	cout << "\n                " << (char)toupper(word[1]) << "\n\n"
	     << "             " <<  (char)toupper(word[2]) << "     " << (char)toupper(word[3]) << "\n\n" 
	     << "\n                " << (char)toupper(word[6]) << "\n\n"
	     << "             " <<  (char)toupper(word[4]) << "     " << (char)toupper(word[5]) << "\n\n" 
	     << "\n                " << (char)toupper(word[0]) << "\n\n";
}

static void
solveit(const char *letters, const char *dpath)
{

	Dict d(dpath, false);
	Solver s(&d);
	DispWordsBuf dsp;

	int n = s.solve(letters);
	if (n != 0) {
		dsp.init(s.words());
		cout << endl << dsp.solBuf() << endl;
	} 
	cout << dsp.nWords() << " words\n\n";
}

static void	
help(void)
{
	cerr << "Spellathon "VERSION"    Copyright (c) 2008 Girish Rao\n\n"
		"spellathon < -p | -s <string> | -i > [ -r min,max ] [ -d /path/to/dict ] [ -c ] [ -n ]\n\n"
		"Main options:\n\n"
		"  -i  Game instructions.\n\n"
		"  -p  Play a new game.\n\n"
	        "  -s  Solve a puzzle.\n"
		"      Enter the central letter followed by\n"
	        "       the remaining letters.\n"
		"      Example: spellathon -s taplerss (central letter is 't').\n\n"
		"Sub options:\n\n"
		"  -r  Play games with #solutions in the range [min,max]\n\n"
		"  -c  Use precomputed dictionary.\n\n"
		"  -d  Use specified dictionary.\n\n"
		"  -n  Don't show score\n\n"
		;
	exit(1);
}

static void	
instr(void)
{
	cout << "Game instructions:\n\n"
		"  o Find all the words of four or more letters from the\n"
		"    letters shown in the puzzle.\n\n  o In making a word, each"
		" letter may be used once only.\n\n  o Each word must contain"
		" the central letter.\n\n  o There should be atleast one seven"
	        " letter word.\n\n\n"
		"Notes:\n\n  - Using hints will greatly reduce your score.\n\n"
		"  - To get the top score/rating you will need to be fast.\n\n"
		"  - Ratings:  Newbie, Average, Good, "
		"Very good, Outstanding, Inhuman.\n\n"
		;
	exit(0);
}
