/*
 * Copyright (c) 2008, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
 * may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <new>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <errno.h>
#include <time.h>
#include <QCoreApplication>
#include <QProgressBar>

#include "main_widget.h"
#include "spellathon_cls.h"
#include "../logic/defines.h"
#include "../logic/dict.h"
#include "../logic/util.h"
#include "../logic/solver.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/score.h"

#ifdef WIN32
/* mingw doesn't have random(3) */
# define random rand
# define srandom srand
#endif

extern MainWidget* gui_h;

namespace Spellathon
{
	
	SpellathonCls::SpellathonCls()
		: dct(0), sol(0), dsp(0)
	{
		minSol_ = MIN_SOL;
		maxSol_ = MAX_SOL;
		ingame = false;
		inprogress = false;
		dictpath = "";
		dict_opened = false;
		precomp = false;

		srandom(time(0));
	}

	void
	SpellathonCls::setSolRange(int min, int max)
	{
		maxSol_ = max;
		minSol_ = min;
	}

	void 
	SpellathonCls::setDoScore(bool e)
	{
		doscore = e;
	}

	void
	SpellathonCls::init(QString dpath, bool pc)
	{
		destroy();

		dictpath = dpath;
		dict_opened = false;
		ingame = false;
		inprogress = false;
		precomp = pc;
		try {
			try {
				dct = new Dict(dictpath.toAscii().data(),
					       precomp);
			} 
			catch (string s) 
			{
				emit noDict(s.c_str());
				destroy();
				return;
			}
	
			try {	
				sol = new Solver(dct);
				dsp = new DispWordsBuf;
			} 
			catch (string s) 
			{
				emit fatalMsg(s.c_str());
				exit(1);
			}
		}
		catch (std::bad_alloc &ba)
		{
			string s = "Can't allocate memory: ";
		       	s += ba.what();
			emit fatalMsg(s.c_str());
			exit(1);
		}

		dict_opened = true;
		emit spellathonReady();	
	}
	
	void
	SpellathonCls::destroy()
	{
		delete dsp;
		delete sol;
		delete dct;
		dsp = 0;
		sol = 0;
		dct = 0;
	}
	
	int 
	SpellathonCls::sidxcmp(const void *a, const void *b)
	{
	        unsigned int (*c)[2] = (unsigned int (*)[2])a;
	        unsigned int (*d)[2] = (unsigned int (*)[2])b;
	
	        if ((*c)[1] == (*d)[1])
	                return 0;
	        if ((*c)[1] < (*d)[1])
	                return -1;
	        else
	                return 1;
	}
	
	void
	SpellathonCls::play_precomp()
	{
		assert(dictpath != "" && precomp);

		inprogress = false;
		ingame = false;

		int i = 0, s = -1, e = -1;

		while (i < dct->pc_nent()) {
			if (dct->precomp_p[i].nsol >= minSol_) {
				s = i;
				break;
			}
			i++;
		}

		while (i < dct->pc_nent()) {
			if (dct->precomp_p[i].nsol > maxSol_) {
				break;
			}
			i++;
		}

		e = i;

		if (s < 0) {
			printMsg("No game", "No suitable game found.\n\n"
			       	 "Choose a different Solution range or use "
				 "a different dictionary.");
		} else {
			int r = randnum(s, e);
			sol->solve(dct->precomp_p[r].sign);
			signFound(dct->precomp_p[r].sign);
		}
	}

	void
	SpellathonCls::play()
	{
		char		sig[MAX_LETRS+1];
		unsigned int 	(*sidx)[2];
		enum	{IDX, RAND};

		assert(dictpath != "" && !precomp);

		inprogress = false;
		ingame = false;

	    sidx = (unsigned int (*)[2])malloc(2 * dct->nent7() * sizeof(int));
		if (sidx == NULL) 
			emit fatalMsg(QString("Can't allocate memory: %1")
				      .arg(strerror(errno)));

		QProgressBar *pbar = new QProgressBar(gui_h->statusOutput);
		pbar->setRange(0, dct->nent7());
		pbar->setTextVisible(false);
		gui_h->statusOutput->insertWidget(0, pbar, 0);

		/* shuffle the indexes. Knuth shuffle? */
		int j = dct->off7();
		if (j < 0)
			goto notfound;

	        for (int i = 0; i < dct->nent7(); i++) {
	                sidx[i][IDX] = j++;
	                sidx[i][RAND] = random();
	        }
	
	        qsort((void*)sidx, dct->nent7(), 2 * sizeof(int), sidxcmp);
	
		unsigned int cidx[MAX_LETRS][2];
	        for (int i = 0; i < MAX_LETRS; i++) {
	                cidx[i][IDX] = i;
	                cidx[i][RAND] = random();
	        }
	
	        qsort((void*)cidx, MAX_LETRS, 2 * sizeof(int), sidxcmp);
	
		/*
		 * find a random seven letter signature from the dict and solve
		 * it and any of the six different central letter combinations.
		 * consider the word playable if MIN_SOL <= solutions <= MAX_SOL
		 *
		 * this technique works well. it's fast, sufficiently
		 * random and covers all the signatures.
		 */
		for (int i = 0; i < dct->nent7(); i++) {

			const char *dsig = (*dct).dp[sidx[i][IDX]].sign;

			int j;
			for (j = 0 ; j < MAX_LETRS; j++)
				sig[j] = dsig[cidx[j][IDX]]; 

			sig[j] = 0;

			int k = MAX_LETRS - 2;
			do {
				if (sol->solve(sig, minSol_, maxSol_) != -1) {
					/* found suitable sign/game */
					gui_h->statusOutput->removeWidget(pbar);
					free(sidx);
					signFound(sig);
					return;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);

			} while (1);

			/* 
			 * this can take some time in pathological cases 
			 * so we need to keep the gui responsive 
			 */	
			if (!(i & ((1<<8) - 1))) {
				pbar->setValue(i);
				QCoreApplication::processEvents();
			}
		}
notfound:
		/* Uh Oh */	
		gui_h->statusOutput->removeWidget(pbar);
		free(sidx);
		printMsg("No game", "No suitable game found.\n\n"
	        	 "Choose a different Solution range or use "
			 "a different dictionary.");
	}

	void
	SpellathonCls::signFound(char *sig)
	{
		ingame = true;
		t = time(0);
	
		emit printPuzzle(sig);

		dsp->init(sol->words());
		emit printWords(dsp->hintBuf());

		emit printWordCnt(QString("Words:  0 / %1 ").arg(dsp->nWords()));
	}
	
	void
	SpellathonCls::hint()
	{
		assert(ingame);
	
		dsp->setHint();
		emit printWords(dsp->hintBuf());
	
		if (dsp->hintLevel() >= 2)
			emit disableHints();
	}
	
	void
	SpellathonCls::finish()
	{
		assert(ingame);
	
		time_t t2 = time(0) - t;

		emit printWords(dsp->solBuf());

		if (doscore) {
			Score sc(dsp->nSetWords(), dsp->nWords(), 
				 dsp->hintLevel(), t2);

			emit printStatus(QString("Time:  %1m%2s      Score:  %3")
					 .arg((int)t2/60)
					 .arg((int)t2%60).arg(sc.score()));

			emit printMsg("Score", 
				      QString("Words:  %1  out of  %2."
					      "\n\nScore:  %3            Rating: %4.        \n")
				      .arg(dsp->nSetWords()).arg(dsp->nWords())
				      .arg(sc.score()).arg(sc.rating().c_str()));
		} else {
			emit printStatus(QString("Time:  %1m%2s")
					 .arg((int)t2/60).arg((int)t2%60));
		}

		ingame = false;
		inprogress = false;
	}
	
	void
	SpellathonCls::solve(QString str)
	{
		assert(dictpath != "" && !ingame);

		int n = sol->solve(str.toAscii().data());
		if (n > 0) {
			dsp->init(sol->words());
	
			emit printWords(dsp->solBuf());
	
			emit printWordCnt(QString("Words:  %1 ").arg(dsp->nWords()));
		} else {
			emit printWordCnt("          ");
			emit printStatus("No Solution.");
		}
	}
	
	void
	SpellathonCls::setWord(QString w)
	{
		assert(ingame);
	
		switch (dsp->setWord(w.toStdString())) {
		case 1:
			inprogress = true;

			emit printWordCnt(QString("Words:  %1 / %2 ")
					  .arg(dsp->nSetWords())
					  .arg(dsp->nWords()));

			if (dsp->nWords() == dsp->nSetWords()) {
				emit finished();
			} else {
				emit printWords(dsp->hintBuf());
				emit wordchkimg(true);
			}
			break;

		case -1:
			emit wordchkimg(false);
			break;

		case 0:
			emit printStatusTemp("Already entered.", 3000);
			break;
		}
			
	}

}

